"""Chat Controller"""
import ujson
import openai
import asyncio
from typing import Any, Union, AsyncIterable

# from fastapi import Request
from langchain.chains import LLMChain
from langchain.callbacks import AsyncIteratorCallbackHandler, get_openai_callback

from promptengineers.core.config.llm import ACCEPTED_OLLAMA_MODELS, ACCEPTED_OPENAI_MODELS, OpenAIModels
from promptengineers.core.interfaces.repos import IUserRepo
from promptengineers.models.message import SystemMessage, UserMessage, AssistantMessage
from promptengineers.repos.user import UserRepo
from promptengineers.llms.services.openai import openai_chat_functions_model
from promptengineers.llms.services.langchain.callbacks import AgentStreamCallbackHandler
from promptengineers.llms.services.langchain.chains import ChainService
from promptengineers.llms.strategies import OllamaStrategy, OpenAIStrategy, ModelContext
from promptengineers.llms.utils import retrieve_chat_messages, retrieve_system_message, get_chat_history, combine_documents
from promptengineers.retrieval.strategies import VectorstoreContext
from promptengineers.stream.utils import token_stream, wrap_done
from promptengineers.core.validations import Validator
from promptengineers.prompts.templates import get_system_template, get_retrieval_template

validator = Validator()

class ChatController:
	def __init__(
		self,
		user_id: str = None,
		# request: Request = None,
		request = None,
		user_repo: IUserRepo = None,
		available_tools: dict[str, Any] = None
	):
		self.request = request
		self.user_id = user_id or getattr(request.state, "user_id", None)
		self.user_repo = user_repo or UserRepo()
		self.available_tools = available_tools or None

	#######################################################
	## Open AI Chat GPT
	#######################################################
	def openai_http_chat(
		self,
		messages,
		model:str,
		temperature: float or int = 0.0,
		stream: bool = False,
	) -> AsyncIterable[str]:
		"""Send a message to the chatbot and yield the response."""
		response = openai.ChatCompletion.create(
			model=model,
			messages=messages,
			temperature=temperature,
			stream=stream
		)
		return response

	async def openai_stream_chat(
		self,
		messages,
		model:str,
		temperature: float or int = 0.0
	) -> AsyncIterable[str]:
		"""Send a message to the chatbot and yield the response."""
		response = openai.ChatCompletion.create(
			model=model,
			messages=messages,
			temperature=temperature,
			stream=True
		)
		for chunk in response:
			## Would also consider gathering data here
			token = chunk['choices'][0]['delta'].get('content', '')
			yield token_stream(token)
		yield token_stream()

	#######################################################
	## Open AI Function Calling
	#######################################################
	def openai_http_function_chat(
		self,
		messages,
		model: str,
		functions: list[str],
		temperature: float or int = 0
	):
		"""Send a message to the chatbot and yield the response."""
		response = openai_chat_functions_model(
			messages=messages,
			model_name=model,
			temperature=temperature,
			streaming=False,
			keys=functions,
		)
		return response

	async def openai_stream_function_chat(
		self,
		messages,
		model: str,
		functions: list[str],
		temperature: float or int = 0
	) -> AsyncIterable[str]:
		"""Send a message to the chatbot and yield the response."""
		response = openai_chat_functions_model(
			messages=messages,
			model_name=model,
			temperature=temperature,
			streaming=True,
			keys=functions,
		)
		for chunk in response:
			token = chunk['choices'][0]['delta'].get('content', '')
			yield token_stream(token)
		yield token_stream()

	##############################################################################
	## Normal Chat
	##############################################################################
	def langchain_http_chat(
		self,
		messages,
		model: str,
		temperature: float or int = 0.0,
	) -> (str, Any):
		# Retrieve the chat messages
		filtered_messages = retrieve_chat_messages(messages)
		# Retrieve the chat history
		chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
		# Retrieve the system message
		system_message = retrieve_system_message(messages)
		# Get Tokens
		api_key = self.user_repo.find_token(self.user_id, 'OPENAI_API_KEY')
		# Check allowed
		if model in ACCEPTED_OPENAI_MODELS:
			model_service = ModelContext(strategy=OpenAIStrategy(api_key=api_key))
		# Construct the model
		model = model_service.chat(
			model_name=model,
			temperature=temperature,
			streaming=False
		)
		query = {'question': filtered_messages[-1], 'chat_history': chat_history}
		with get_openai_callback() as cb:
			# Retrieve the conversation
			chain = LLMChain(llm=model, prompt=get_system_template(system_message))
			# Begin a task that runs in the background.
			response = chain.run(query)
		return response, cb

	##############################################################################
	## Langchain Agent Chat
	##############################################################################
	async def langchain_http_agent_chat(
		self,
		messages,
		model: str,
		tools,
		temperature: float or int = 0.0,
	) -> (str, Any):
		"""Send a message to the chatbot and yield the response."""
		filtered_messages = retrieve_chat_messages(messages)
		# Retrieve the chat history
		chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
		# Retrieve the system message
		system_message = retrieve_system_message(messages)
		# Attach the application user id to the system message
		system_message = system_message + '\n' + "USER_ID=" + str(self.user_id)
		# Get Tokens
		api_key = self.user_repo.find_token(self.user_id, 'OPENAI_API_KEY')
		# Create the model
		if model in ACCEPTED_OPENAI_MODELS:
			model_service = ModelContext(strategy=OpenAIStrategy(api_key=api_key))
		else:
			raise NotImplementedError(f"Model {model} not implemented")

		model = model_service.chat(
			model_name=model,
			temperature=temperature,
			streaming=False
		)
		with get_openai_callback() as cb:
			# Retrieve the conversation
			chain = ChainService(model).agent_with_tools(system_message=system_message,
														chat_history=chat_history,
														tools=tools)
			# Begin a task that runs in the background.
			response = await chain.ainvoke(filtered_messages[-1])
		return response, cb

	##############################################################################
	## Langchain Vectorstore Chat
	##############################################################################
	def langchain_http_vectorstore_chat(
		self,
		messages: list[Union[SystemMessage, UserMessage, AssistantMessage]] = None,
		model: str = OpenAIModels.GPT_3_5_TURBO_16K.value,
		temperature: float or int = 0.9,
		vectorstore: VectorstoreContext = None,
	) -> (str, Any):
		"""Send a message to the chatbot and yield the response."""
		filtered_messages = retrieve_chat_messages(messages)
		# Retrieve the chat history
		chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2])) ## TODO: Fix this
		# Retrieve the system message
		system_message = retrieve_system_message(messages)
		# Get Tokens
		api_key = self.user_repo.find_token(self.user_id, 'OPENAI_API_KEY')
		# Create the model
		if model in ACCEPTED_OPENAI_MODELS:
			model_service = ModelContext(strategy=OpenAIStrategy(api_key=api_key))
		else:
			raise NotImplementedError(f"Model {model} not implemented")

		model = model_service.chat(
			model_name=model,
			temperature=temperature,
			streaming=False
		)
		with get_openai_callback() as cb:
			# Retrieve the conversation
			chain = ChainService(model).conversation_retrieval(
				system_message=system_message, vectorstore=vectorstore, chat_history=chat_history
			)
			# Begin a task that runs in the background.
			response = chain(filtered_messages[-1], return_only_outputs=True)
		return response, cb

	##############################################################################
	## Langchain Chat Stream
	##############################################################################
	async def langchain_stream_chat(
		self,
		messages,
		model:str,
		temperature: float or int = 0.0,
		stream: bool = True,
	) -> AsyncIterable[str]:
		"""Send a message to the chatbot and yield the response."""
		filtered_messages = retrieve_chat_messages(messages)
		# Retrieve the chat history
		chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
		# Retrieve the system message
		system_message = retrieve_system_message(messages)

		# Create the model
		if model in ACCEPTED_OPENAI_MODELS:
			# Get Tokens
			api_key = self.user_repo.find_token(self.user_id, 'OPENAI_API_KEY')
			model_service = ModelContext(strategy=OpenAIStrategy(api_key=api_key))
			callback = AsyncIteratorCallbackHandler()
			model = model_service.chat(
				model_name=model,
				temperature=temperature,
				streaming=stream,
				callbacks=[callback]
			)
			query = {'question': filtered_messages[-1], 'chat_history': chat_history}
			# Retrieve the conversation
			chain = LLMChain(llm=model, prompt=get_system_template(system_message))
			# Begin a task that runs in the background.
			task = asyncio.create_task(wrap_done(
				chain.acall(query),
				callback.done),
			)
			# Yield the tokens as they come in.
			async for token in callback.aiter():
				yield token_stream(token)
			yield token_stream()
			await task
		elif model in ACCEPTED_OLLAMA_MODELS:
			# Get Tokens
			base_url = self.user_repo.find_token(self.user_id, 'OLLAMA_BASE_URL')
			if base_url:
				strategy = OllamaStrategy(base_url=base_url)
			else:
				strategy = OllamaStrategy()
			model_service = ModelContext(strategy=strategy)
			callback = AgentStreamCallbackHandler()
			llm = model_service.chat(
				model_name=model,
				temperature=temperature,
				streaming=stream,
				callbacks=[callback]
			)
			template = get_system_template(system_message)
			prompt = template.format(
				context=get_chat_history(chat_history),
				question=filtered_messages[-1]
			)
			# Yield the tokens as they come in.
			for token in llm._stream(prompt):
				yield token_stream(token.text)
			yield token_stream()

	#######################################################
	## Langchain Agent Stream Chat
	#######################################################
	async def langchain_stream_agent_chat(
		self,
		messages,
		model: str,
		tools,
		temperature: float or int = 0.0,
	):
		"""Send a message to the chatbot and yield the response."""
		filtered_messages = retrieve_chat_messages(messages)
		# Retrieve the chat history
		chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
		# Retrieve the system message
		system_message = retrieve_system_message(messages)
		# Attach the application user id to the system message
		system_message = system_message + '\n' + "USER_ID=" + str(self.user_id)
		# Get Tokens
		api_key = self.user_repo.find_token(self.user_id, 'OPENAI_API_KEY')
		# Create the model
		if model in ACCEPTED_OPENAI_MODELS:
			model_service = ModelContext(strategy=OpenAIStrategy(api_key=api_key))
		else:
			raise NotImplementedError(f"Model {model} not implemented")

		callback = AgentStreamCallbackHandler()
		model = model_service.chat(
			model_name=model,
			temperature=temperature,
			streaming=True,
			callbacks=[callback]
		)
		query = {'input': filtered_messages[-1], 'chat_history': chat_history}
		# tools = load_tools(tools, llm=model)
		agent_executor = ChainService(model).agent_with_tools(system_message=system_message,
															chat_history=chat_history,
															tools=tools,
															callbacks=[callback])
		runnable = agent_executor.astream_log(query)
		async for chunk in runnable:
			operation = chunk.ops[0]['value']
			if operation:
				if type(operation) == str:
					filled_chunk = operation
					if filled_chunk:
						yield token_stream(filled_chunk)
				else:
					generations = operation.get('generations', False)
					if generations:
						function_call = generations[0][0].get('message', {}).additional_kwargs.get('function_call', {})
						tool = function_call.get('name', None)
						if tool:
							yield token_stream(tool, 'tool')
						args = function_call.get('arguments', None)
						if args:
							if type(args) == str:
								tool_args = ujson.loads(args)
							else:
								tool_args =  ujson.loads(args)['__arg1']

							yield token_stream(f"Invoking: `{tool}` with `{tool_args}`", 'log')
		yield token_stream()

	#######################################################
	## Vectorstore
	#######################################################
	async def langchain_stream_vectorstore_chat(
		self,
		messages: list[str],
		model: str,
		temperature: float or int = 0.9,
		vectorstore: VectorstoreContext = None,
	) -> AsyncIterable[str]:
		"""Send a message to the chatbot and yield the response."""
		filtered_messages = retrieve_chat_messages(messages)
		# Retrieve the chat history
		chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
		# Retrieve the system message
		system_message = retrieve_system_message(messages)
		# Get Tokens
		api_key = self.user_repo.find_token(self.user_id, 'OPENAI_API_KEY')
		# Create the callback
		callback = AgentStreamCallbackHandler()
		# Create the model
		if model in ACCEPTED_OPENAI_MODELS:
			model_service = ModelContext(strategy=OpenAIStrategy(api_key=api_key))
			llm = model_service.chat(
				model_name=model,
				temperature=temperature,
				streaming=True,
				callbacks=[callback]
			)
			# Retrieve the conversation
			qa_chain = ChainService(llm).conversation_retrieval(
				system_message=system_message,
				chat_history=chat_history,
				vectorstore=vectorstore,
				callbacks=[callback]
			)
			runnable = qa_chain.astream_log(filtered_messages[-1])
			docs_processed = False
			async for chunk in runnable:
				operation = chunk.ops[0]['value']
				# print(operation)
				async for chunk in runnable:
					operation = chunk.ops[0]['value']
					return_output = False

					if isinstance(operation, dict):
						docs = operation.get('documents')
						if docs:
							for doc in docs:
								yield token_stream({
									'page_content': doc.page_content,
									'metadata': doc.metadata,
								}, 'doc')
							docs_processed = True  # Set this flag when docs are processed

					# Check if docs have been processed at least once
					if docs_processed:
						return_output = True

					if operation and isinstance(operation, str) and return_output:
						yield token_stream(operation)
			yield token_stream()

		elif model in ACCEPTED_OLLAMA_MODELS:
			base_url = self.user_repo.find_token(self.user_id, 'OLLAMA_BASE_URL')
			if base_url:
				strategy = OllamaStrategy(base_url=base_url)
			else:
				strategy = OllamaStrategy()
			model_service = ModelContext(strategy=strategy)
			callback = AgentStreamCallbackHandler()
			llm = model_service.chat(
				model_name=model,
				temperature=temperature,
				streaming=True,
				callbacks=[callback]
			)
			question = filtered_messages[-1]
			template = get_retrieval_template(system_message)
			history = get_chat_history(chat_history)
			documents = vectorstore.similarity_search(question)
			context = '\n'.join([doc.page_content for doc in documents])
			prompt = template.format(
				chat_history=history,
				context=context,
				question=question
			)
			# Yield the tokens as they come in.
			for token in llm._stream(prompt):
				yield token_stream(token.text)
			yield token_stream()
		else:
			raise NotImplementedError(f"Model {model} not implemented")



from typing import List
import os
import time
import pickle
import tempfile
from queue import Queue
from concurrent.futures import ThreadPoolExecutor

from fastapi import HTTPException, UploadFile, File, Request

from promptengineers.core.config.loaders import FileLoaderType
from promptengineers.retrieval.factories.embedding import EmbeddingFactory
from promptengineers.retrieval.factories.loader import LoaderFactory
from promptengineers.core.interfaces.repos import IUserRepo
from promptengineers.repos.user import UserRepo
from promptengineers.retrieval.services.redis import RedisService
from promptengineers.retrieval.services.pinecone import PineconeService
from promptengineers.storage.services import StorageService
from promptengineers.core.utils import logger
from promptengineers.core.validations import Validator
from promptengineers.retrieval.utils import create_faiss_vectorstore
from promptengineers.core.config.llm import OllamaModels

validator = Validator()
user_repo = UserRepo()

##############################################################
### Process File
##############################################################
def process_file(
	index_name: str,
	tmpdirname: str,
	file: UploadFile,
	embeddings,
	pinecone_service: PineconeService,
):
	try:
		file_path = os.path.join(tmpdirname, file.filename)
		_, file_extension = os.path.splitext(file.filename)
		file_extension_cleaned = file_extension.replace('.', '')

		# Check if the file extension is a valid type
		if file_extension_cleaned not in [e.value for e in FileLoaderType]:
			raise HTTPException(status_code=400, detail=f"Invalid file type: {file_extension_cleaned}")

		with open(file_path, 'wb') as f:
			f.write(file.file.read())  # write the file to the temporary directory
		doc_loader = LoaderFactory.create_loader(file_extension_cleaned, {'file_path': file_path})

		pinecone_service.from_documents([doc_loader], embeddings, namespace=index_name)
	except HTTPException:
		raise  # Re-raise the HTTPException without modifying it
	except Exception as e:
		raise HTTPException(status_code=400, detail=str(e)) from e

##############################################################
### Accumulate Files
##############################################################
def accumulate_files(files, user_id, tokens):
	collected_files = []
	for name in files:
		try:
			s3 = StorageService(tokens.get('ACCESS_KEY_ID'), tokens.get('ACCESS_SECRET_KEY'))
			file = s3.retrieve_file(
				path=f'users/{user_id}/files/{name}',
				bucket=tokens.get('BUCKET'),
			)
			collected_files.append(file)
		except Exception as err:
			logger.error(err)
			raise HTTPException(
				status_code=404, detail=f"File Not Found in S3: {str(name)}"
			) from err
	filtered_lst = [x for x in collected_files if x is not None]
	return filtered_lst

##############################################################
### Accumulate Loaders
##############################################################
def accumulate_loaders(body, files=None, tmpdirname=None):
	loaders = []
	for loader in dict(body).get('loaders', []):
		try:
			# First, try accessing as if loader is a dictionary
			loader_type = loader['type']
		except TypeError:
			# If it fails, then access it as an attribute
			loader_type = loader.type
		loader_data = {}
		if loader_type == 'copy':
			try:
				loader_data = {'text': loader['text']}
			except TypeError:
				loader_data = {'text': loader.text}
		elif loader_type == 'yt':
			try:
				loader_data = {'ytId': loader['ytId']}
			except TypeError:
				loader_data = {'ytId': loader.ytId}
		elif loader_type in ['ethereum', 'polygon']:
			try:
				loader_data = {'contract_address': loader['contract_address']}
			except TypeError:
				loader_data = {'contract_address': loader.contract_address}
		else:
			try:
				loader_data = {'urls': loader['urls']}
			except TypeError:
				loader_data = {'urls': loader.urls}


		doc_loader = LoaderFactory.create_loader(loader_type,  loader_data)
		loaders.append(doc_loader)

	if tmpdirname:
		try:
			for file_body, name in zip(files, body.files):
				file_path = os.path.join(tmpdirname, name)
				with open(file_path, 'wb') as file:
					file.write(file_body.read())  # write the file to the temporary directory
				filename = file_path.split('/')[-1]
				loader = os.path.splitext(filename)[1][1:]
				doc_loader = LoaderFactory.create_loader(
					loader,
					{'file_path': file_path}
				)
				loaders.append(doc_loader)
		except Exception as err:
			logger.error(err)
			raise HTTPException(
				status_code=500,
				detail=f"Error creating loaders in tmp directory: {str(err)}"
			) from err
	return loaders

##############################################################
### Create a FAISS Vectorstore
##############################################################
def faiss_vectorstore(loaders, tmpdirname, user_id, name, tokens):
	documents = []
	for loader in loaders:
		documents.extend(loader.load())
	logger.info("[main.load_vectorstore_from_file] Loaders: %s", str(len(loaders)))
	logger.info("[main.load_vectorstore_from_file] Documents: %s", str(len(documents)))
	vectorstore = create_faiss_vectorstore(documents)
	## Save Vectorstore to tmp directory
	temp_file_path = os.path.join(tmpdirname, f'{int(time.time())}-{name}.pkl')
	try:
		## Write to temp file
		with open(temp_file_path, "wb") as file:
			pickle.dump(vectorstore, file)

		s3 = StorageService(tokens.get('ACCESS_KEY_ID'), tokens.get('ACCESS_SECRET_KEY'))
		# Save to S3
		file = s3.upload_file(
			file_name=temp_file_path,
			bucket=tokens.get('BUCKET'),
			object_name=f'users/{user_id}/vectorstores/{name}.pkl'
		)
	except Exception as err:
		raise HTTPException(
			status_code=500,
			detail=f"Error saving vectorstore in tmp directory: {str(err)}"
		) from err
	finally:
		with os.scandir(tmpdirname) as entries:
			for entry in entries:
				os.remove(entry.path)


class VectorSearchController:
	def __init__(self, request: Request = None, user_repo: IUserRepo = None):
		# Initialize any necessary variables or objects here
		self.request = request
		self.user_id = getattr(request.state, "user_id", None)
		self.user_repo = user_repo or UserRepo()

	##############################################################
	### Create multi loader vectorstore
	##############################################################
	async def create_multi_loader_vectorstore(self, body):
		"""Create a vectorstore from multiple loaders."""
		req_body = dict(body)
		passed_file_names = req_body.get('files', [])
		pinecone_keys = ['PINECONE_API_KEY', 'PINECONE_ENV', 'PINECONE_INDEX', 'OPENAI_API_KEY']
		redis_keys = ['REDIS_URL', 'OPENAI_API_KEY']
		if not passed_file_names:
			if req_body.get('provider') == 'pinecone':
				tokens = self.user_repo.find_token(self.user_id, pinecone_keys)
				embeddings = EmbeddingFactory(req_body.get('embedding'), tokens.get('OPENAI_API_KEY'))
				validator.validate_api_keys(tokens, pinecone_keys)
				pinecone_service = PineconeService(
					api_key=tokens.get('PINECONE_API_KEY'),
					env=tokens.get('PINECONE_ENV'),
					index_name=tokens.get('PINECONE_INDEX'),
				)
				loaders = accumulate_loaders(body)
				pinecone_service.from_documents(
					loaders,
					embeddings(),
					namespace=dict(body).get('index_name')
				)

			if req_body.get('provider') == 'redis':
				tokens = self.user_repo.find_token(self.user_id, redis_keys)
				validator.validate_api_keys(tokens, redis_keys)
				embeddings = EmbeddingFactory(req_body.get('embedding'), tokens.get('OPENAI_API_KEY'))
				redis_service = RedisService(
					embeddings=embeddings(),
					redis_url=tokens.get('REDIS_URL'),
					index_name=req_body.get('index_name'),
				)
				loaders = accumulate_loaders(body)
				redis_service.add_docs(loaders)
		else:
			aws_keys = ['ACCESS_KEY_ID', 'ACCESS_SECRET_KEY', 'BUCKET']
			tokens = self.user_repo.find_token(self.user_id, [*pinecone_keys, *aws_keys])
			validator.validate_api_keys(tokens, [*pinecone_keys, *aws_keys])
			# Accumulate Files
			files = accumulate_files(passed_file_names, self.user_id, tokens)

			# Create a temporary directory
			with tempfile.TemporaryDirectory() as tmpdirname:

				# Your logic here to save uploaded files to tmpdirname
				loaders = accumulate_loaders(body, files, tmpdirname)

				if req_body.get('provider') == 'faiss':
					faiss_vectorstore(loaders, tmpdirname, self.user_id, req_body.get('index_name'), tokens)

				if req_body.get('provider') == 'pinecone':
					embeddings = EmbeddingFactory(req_body.get('embedding'), tokens.get('OPENAI_API_KEY'))
					pinecone_service = PineconeService(
						api_key=tokens.get('PINECONE_API_KEY'),
						env=tokens.get('PINECONE_ENV'),
						index_name=tokens.get('PINECONE_INDEX'),
					)
					pinecone_service.from_documents(loaders, embeddings, namespace=req_body.get('index_name'))

				if req_body.get('provider') == 'redis':
					tokens = self.user_repo.find_token(self.user_id, redis_keys)
					validator.validate_api_keys(tokens, redis_keys)
					redis_service = RedisService(
						openai_api_key=tokens.get('OPENAI_API_KEY'),
						redis_url=tokens.get('REDIS_URL'),
						index_name=req_body.get('index_name'),
					)
					loaders = accumulate_loaders(body)
					redis_service.add_docs(loaders)

	##############################################################
	### Create a Vectorstore from files
	##############################################################
	async def create_vectorstore_from_files(
		self,
		provider: str,
		index_name: str,
		embedding: str,
		files: List[UploadFile] = File(...),
		threaded: bool = True,
	):
		"""Create a vectorstore from files."""
		# Validate the file extensions before processing
		for file in files:
			_, file_extension = os.path.splitext(file.filename)
			file_extension_cleaned = file_extension.replace('.', '')

			# Check if the file extension is a valid type
			if file_extension_cleaned not in [e.value for e in FileLoaderType]:
				raise HTTPException(status_code=400, detail=f"Invalid file type: {file_extension_cleaned}")

		## Get Tokens
		if provider == 'pinecone':
			keys = ['PINECONE_API_KEY', 'PINECONE_ENV', 'PINECONE_INDEX', 'OPENAI_API_KEY']
			tokens = self.user_repo.find_token(self.user_id, keys)
			## Check for token, else throw error
			validator.validate_api_keys(tokens, keys)

			## Get Embeddings and Pinecone Service
			embeddings = EmbeddingFactory(embedding, tokens.get('OPENAI_API_KEY'))
			pinecone_service = PineconeService(
				api_key=tokens.get('PINECONE_API_KEY'),
				env=tokens.get('PINECONE_ENV'),
				index_name=tokens.get('PINECONE_INDEX'),
			)

		## Create a temporary directory
		with tempfile.TemporaryDirectory() as tmpdirname:
			# Your logic here to save uploaded files to tmpdirname
			if threaded:
				file_queue = Queue()
				for file in files:
					file_queue.put(file)

				def worker():
					while not file_queue.empty():
						file = file_queue.get()
						try:
							process_file(index_name, tmpdirname, file, embeddings(), pinecone_service)
						finally:
							file_queue.task_done()

				with ThreadPoolExecutor() as executor:
					num_workers = min(len(files), os.cpu_count())  # You can adjust the number of workers
					for _ in range(num_workers):
						executor.submit(worker)

				file_queue.join()  # Wait for all files to be processed
			else:
				for file in files:
					process_file(index_name, tmpdirname, file, embeddings(), pinecone_service)


	##############################################################
	### Retrieve Pinecone Vectorstores
	##############################################################
	def retrieve_pinecone_vectorstores(self):
		## Get Tokens
		keys = ['PINECONE_API_KEY', 'PINECONE_ENV', 'PINECONE_INDEX']
		tokens = self.user_repo.find_token(self.user_id, keys)
		## Check for token, else throw error
		validator.validate_api_keys(tokens, keys)
		## Get Vectorstores
		pinecone_service = PineconeService(
			api_key=tokens.get('PINECONE_API_KEY'),
			env=tokens.get('PINECONE_ENV'),
			index_name=tokens.get('PINECONE_INDEX'),
		)
		index_stats = pinecone_service.describe_index_stats()
		namespaces = index_stats.get('namespaces')

		return {
			'vectorstores': list(namespaces.keys()),
		}

	##############################################################
	### Delete Pinecone Vectorstore
	##############################################################
	def delete_pinecone_vectorstore(self, prefix: str):
		## Get Tokens
		keys = ['PINECONE_API_KEY', 'PINECONE_ENV', 'PINECONE_INDEX']
		tokens = self.user_repo.find_token(self.user_id, keys)
		## Check for token, else throw error
		validator.validate_api_keys(tokens, keys)
		## Delete Vectorstore
		pinecone_service = PineconeService(
			api_key=tokens.get('PINECONE_API_KEY'),
			env=tokens.get('PINECONE_ENV'),
			index_name=tokens.get('PINECONE_INDEX'),
		)
		deleted = pinecone_service.delete(namespace=prefix)
		if deleted:
			return True
		else:
			return False

	##############################################################
	### Retrieve Pinecone Vectorstores
	##############################################################
	def retrieve_redis_vectorstores(self):
		## Get Tokens
		keys = ['REDIS_URL']
		tokens = self.user_repo.find_token(self.user_id, keys)
		## Check for token, else throw error
		validator.validate_api_keys(tokens, keys)
		## Get Vectorstores
		redis_service = RedisService(
			redis_url=tokens.get('REDIS_URL')
		)
		index_stats = redis_service.list_indexes()

		return {
			'vectorstores': list(index_stats),
		}

	##############################################################
	### Delete Pinecone Vectorstore
	##############################################################
	def delete_redis_vectorstore(self, prefix: str):
		## Get Tokens
		keys = ['REDIS_URL']
		tokens = self.user_repo.find_token(self.user_id, keys)
		## Check for token, else throw error
		validator.validate_api_keys(tokens, keys)
		## Delete Vectorstore
		redis_service = RedisService(
			redis_url=tokens.get('REDIS_URL'),
			index_name=prefix,
		)
		deleted = redis_service.delete(index_name=prefix)
		if deleted:
			return True
		else:
			return False


from typing import List
import traceback
import ujson

from fastapi import (APIRouter, Depends, Request, Form, status,
                    Response, UploadFile, File, HTTPException)

from promptengineers.core.exceptions import ValidationException
from promptengineers.models.request import RequestDataLoader, RequestMultiLoader
from promptengineers.models.response import (ResponseFileLoader, ResponseCreateVectorStore,
									ResponseListPineconeVectorStores)
from promptengineers.fastapi.controllers import VectorSearchController, AuthController
from promptengineers.core.utils import logger

TAG = "Retrieval"
router = APIRouter()
auth_controller = AuthController()

def get_controller(request: Request) -> VectorSearchController:
	return VectorSearchController(request=request)

#################################################
# Create vectorstore from files
#################################################
@router.post(
	"/vectorstores",
	response_model=ResponseCreateVectorStore,
	tags=[TAG]
)
async def create_vectorstore(
	body: RequestDataLoader,
	controller: VectorSearchController = Depends(get_controller),
):
	"""File Loader endpoint."""
	logger.debug('[POST /vectorstores] Body: %s', str(body))
	try:
		await controller.create_multi_loader_vectorstore(body)

		## Format Response
		data = ujson.dumps({
			"message": 'Vectorstore Created!',
			'vectorstore': body.index_name,
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=201
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("HTTPException: %s", err.detail, stack_info=True)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.vectorstores.create_vectorstore]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Create vectorstore from files
#################################################
@router.post(
	"/vectorstores/file",
	response_model=ResponseFileLoader,
	tags=[TAG]
)
async def create_vectorstore_from_file(
	index_name: str = Form(...),
	provider: str = Form(...),
	files: List[UploadFile] = File(...),
	controller: VectorSearchController = Depends(get_controller),
):
	"""File Loader endpoint."""
	try:
		await controller.create_vectorstore_from_files(
			provider,
			index_name,
			files
		)

		## Format Response
		data = ujson.dumps({
			"message": 'Vectorstore Created!',
			'vectorstore': index_name,
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=201
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("HTTPException: %s", err.detail, stack_info=True)
		raise
	except Exception as err:
		logger.error("%s", err, stack_info=True)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err


#################################################
# Create vectorstore
#################################################
@router.post(
	"/vectorstores/multi",
	response_model=ResponseCreateVectorStore,
	tags=[TAG],
	# include_in_schema=False # TODO: Needs some work
)
async def create_vectorstore_from_multiple_sources(
	request: Request,
	body: RequestMultiLoader,
	controller: VectorSearchController = Depends(get_controller),
):
	# Get Body
	body = await request.json()
	logger.debug('[POST /vectorstores/multi] Body: %s', str(body))

	try:
		await controller.create_multi_loader_vectorstore(body)
		## Format Response
		data = ujson.dumps({
			'message': 'Vectorstore Created!',
			'data': body
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=201
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("%s", err.detail, stack_info=True)
		raise
	except Exception as err:
		logger.error("%s", err, stack_info=True)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err


######################################
# Retrieve Vector Stores
######################################
@router.get(
	"/vectorstores/pinecone",
	response_model=ResponseListPineconeVectorStores,
	tags=[TAG]
)
async def list_pinecone_vectorstores(controller: VectorSearchController = Depends(get_controller)):
	try:
		result = controller.retrieve_pinecone_vectorstores()
		# Format Response
		data = ujson.dumps({
			**result
		})
		logger.debug("List Pinecone Vectorstores: %s", data)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("HTTPException: %s", err.detail)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.retrieval.list_pinecone_vectorstores]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

######################################
# Delete Vector Store
######################################
@router.delete(
	"/vectorstores/pinecone",
	status_code=status.HTTP_204_NO_CONTENT,
	tags=[TAG]
)
async def delete_pinecone_vectorstore(
	prefix: str or None = None,
	controller: VectorSearchController = Depends(get_controller)
):
	try:
		controller.delete_pinecone_vectorstore(prefix)
		return Response(status_code=204)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except Exception as err:
		raise HTTPException(status_code=404, detail=str(err)) from err

######################################
# Retrieve Vector Stores
######################################
@router.get(
	"/vectorstores/redis",
	response_model=ResponseListPineconeVectorStores,
	tags=[TAG]
)
async def list_redis_vectorstores(controller: VectorSearchController = Depends(get_controller)):
	try:
		result = controller.retrieve_redis_vectorstores()
		# Format Response
		data = ujson.dumps({
			**result
		})
		logger.debug("List Redis Vectorstores: %s", data)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("HTTPException: %s", err.detail)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.vectorstores.list_redis_vectorstores]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

######################################
# Delete Vector Store
######################################
@router.delete(
	"/vectorstores/redis",
	status_code=status.HTTP_204_NO_CONTENT,
	tags=[TAG]
)
async def delete_redis_vectorstore(
	prefix: str or None = None,
	controller: VectorSearchController = Depends(get_controller)
):
	try:
		controller.delete_redis_vectorstore(prefix)
		return Response(status_code=204)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except Exception as err:
		raise HTTPException(status_code=404, detail=str(err)) from err


"""Chat History Request Models"""
from typing import Optional
from pydantic import BaseModel

HISTORY = {
	"title": "World Series 2001 Chat History",
	"messages": [
		{"role": "system", "content": "You are a helpful assistant."},
		{"role": "user", "content": 'Who won the 2001 world series?'},
		{
			"role": "assistant", 
			"content": 'The arizona diamondbacks won the 2001 world series.',
			"actions": []
		},
	],
	"setting": "abcd147a12abcdefd9512345",
	"tags": ["baseball", "world series", "2001"],
}

HISTORY_DOCUMENT = {
	"_id": "653e147a126c8e67d951fd20",
	"title": "World Series 2001 Chat History",
	"messages": [
		{"role": "system", "content": "You are a helpful assistant."},
		{"role": "user", "content": 'Who won the 2001 world series?'},
		{
			"role": "assistant", 
			"content": 'The arizona diamondbacks won the 2001 world series.',
			"actions": []
		},
	],
	"setting": "abcd147a12abcdefd9512345",
	"tags": ["baseball", "world series", "2001"],
	"created_at": 1698523723,
	"updated_at": 1698562747
}

class History(BaseModel): # pylint: disable=too-few-public-methods
	"""A message to send to the chatbot."""
	messages: Optional[list]
	title: Optional[str] = None
	setting: Optional[str] = None
	tags: Optional[list[str]] = None

	class Config: # pylint: disable=too-few-public-methods
		"""A message to send to the chatbot."""
		json_schema_extra = {
			"example": HISTORY
		}

class HistoryDocument(History): # pylint: disable=too-few-public-methods
	"""A message to send to the chatbot."""
	_id: str
	created_at: int
	updated_at: int

	class Config: # pylint: disable=too-few-public-methods
		"""Message history document"""
		json_schema_extra = {
			"example": HISTORY_DOCUMENT
		}

class HistoryIndex(HistoryDocument): # pylint: disable=too-few-public-methods
	"""A message to send to the chatbot."""
	histories: list[History]

	class Config: # pylint: disable=too-few-public-methods
		"""A message to send to the chatbot."""
		json_schema_extra = {
			"example": {
				"histories": [HISTORY_DOCUMENT]
			}
		}



from typing import Any, Dict, List, Optional

from langchain.callbacks.streaming_aiter import AsyncIteratorCallbackHandler
from langchain.schema import LLMResult

DEFAULT_ANSWER_PREFIX_TOKENS = ["Final", "Answer", ":"]

class AgentStreamCallbackHandler(AsyncIteratorCallbackHandler):
	"""
	Callback handler that returns an async iterator.
	Only the final output of the agent will be iterated.
	"""
	def __init__(
		self,
		*,
		answer_prefix_tokens: Optional[List[str]] = None,
		strip_tokens: bool = True,
		stream_prefix: bool = False,
	) -> None:
		"""Instantiate AsyncFinalIteratorCallbackHandler.

		Args:
			answer_prefix_tokens: Token sequence that prefixes the answer.
				Default is ["Final", "Answer", ":"]
			strip_tokens: Ignore white spaces and new lines when comparing
				answer_prefix_tokens to last tokens? (to determine if answer has been
				reached)
			stream_prefix: Should answer prefix itself also be streamed?
		"""
		super().__init__()
		if answer_prefix_tokens is None:
			self.answer_prefix_tokens = DEFAULT_ANSWER_PREFIX_TOKENS
		else:
			self.answer_prefix_tokens = answer_prefix_tokens
		if strip_tokens:
			self.answer_prefix_tokens_stripped = [
				token.strip() for token in self.answer_prefix_tokens
			]
		else:
			self.answer_prefix_tokens_stripped = self.answer_prefix_tokens
		self.last_tokens = [""] * len(self.answer_prefix_tokens)
		self.last_tokens_stripped = [""] * len(self.answer_prefix_tokens)
		self.strip_tokens = strip_tokens
		self.stream_prefix = stream_prefix
		self.answer_reached = False

	def append_to_last_tokens(self, token: str) -> None:
		self.last_tokens.append(token)
		self.last_tokens_stripped.append(token.strip())
		if len(self.last_tokens) > len(self.answer_prefix_tokens):
			self.last_tokens.pop(0)
			self.last_tokens_stripped.pop(0)

	def check_if_answer_reached(self) -> bool:
		if self.strip_tokens:
			return self.last_tokens_stripped == self.answer_prefix_tokens_stripped
		else:
			return self.last_tokens == self.answer_prefix_tokens

	async def on_llm_start(
		self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
	) -> None:
		# If two calls are made in a row, this resets the state
		self.done.clear()
		self.answer_reached = False

	async def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
		if self.answer_reached:
			self.done.set()

	async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
		if token:
			self.queue.put_nowait(token)


"""Service for removing PII from text."""
import os

from redis import Redis as RedisClient
from langchain.vectorstores.redis import Redis
from langchain.embeddings.openai import OpenAIEmbeddings

from promptengineers.retrieval.utils import split_docs

class RedisService:
	def __init__(
			self, 
			redis_url: str, 
			index_name: str = None, 
			embeddings = None, 
			openai_api_key: str = None
	):
		self.redis_url = redis_url
		self.index_name = index_name
		self.embeddings = OpenAIEmbeddings(openai_api_key=openai_api_key) if openai_api_key else embeddings
		self.client = Redis(
			redis_url=self.redis_url,
			index_name=self.index_name,
			embedding=self.embeddings
		)

	def add_docs(
		self,
		loaders,
		chunk_size: int = 1000,
		chunk_overlap: int = 100,
	):
		docs = []
		for loader in loaders:
			docs.extend(loader.load())

		## TODO: Find out why this is needed, wouldn't work before without.
		os.environ['REDIS_URL'] = self.redis_url

		return self.client.add_documents(split_docs(docs, chunk_size, chunk_overlap))

	#############################################################
	## Retrieve Vectors from Existing Index
	#############################################################
	def from_existing(
		self,
		schema: dict = {"page_content": "TEXT", "metadata": "HASH"},
	):
		return self.client.from_existing_index(
			index_name=self.index_name,
			embedding=self.embeddings,
			redis_url=self.redis_url,
			schema=schema
		)

	#############################################################
	## List Vectorstores
	#############################################################
	def list_indexes(
		self
	):
		## TODO: Find out why this is needed, wouldn't work before without.
		redis_client = RedisClient.from_url(self.redis_url)
		indexes = redis_client.execute_command('FT._LIST')
		return [result.decode('utf-8') for result in indexes]

	#############################################################
	## Delete Vectors
	#############################################################
	def delete(
		self,
		index_name: str
	):
		os.environ['REDIS_URL'] = self.redis_url
		dropped = self.client.drop_index(index_name=index_name, delete_documents=True)
		if dropped:
			return True
		else:
			return False


"""Utils for error handling."""

def token_not_found_message(available_tokens: dict, required_keys: list[str]):
    if available_tokens is None:
        return ', '.join(required_keys) + " not found in Env settings, please add to Account Settings -> Env and try again."
    missing_keys = [key for key in required_keys if key not in available_tokens]
    if not missing_keys:
        return ""
    return ', '.join(missing_keys) + " not found in Env settings, please add to Account Settings -> Env and try again."


"""Function Service"""""
import ujson

from promptengineers.core.config.functions import AVAILABLE_FUNCTIONS

class FunctionFactory:
    """Function Type Factory"""
    def get_result(self, fn_type, response_message):
        """Get Result"""
        function_args = ujson.loads(response_message["function_call"]["arguments"])

        if fn_type == "get_word_length":
            function_to_call = AVAILABLE_FUNCTIONS[fn_type]
            return function_to_call(
				word=function_args.get("word"),
			)
        raise ValueError("Invalid Function type.")




from typing import Any
# from promptengineers.core.validations import Validator
from promptengineers.core.config.llm import ACCEPTED_OLLAMA_MODELS, ACCEPTED_OPENAI_MODELS
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.embeddings import OllamaEmbeddings

# validator = Validator()

class EmbeddingFactory:
	def __init__(self, llm: str, token: str = None, base_url: str = None):
		self.llm = llm
		self.token = token
		self.base_url = base_url

	def __call__(self) -> Any:
		if self.llm in ACCEPTED_OPENAI_MODELS:
			return OpenAIEmbeddings(openai_api_key=self.token)
		elif self.llm in ACCEPTED_OLLAMA_MODELS:
			return OllamaEmbeddings(base_url=self.base_url or 'http://127.0.0.1:11434')
		else:
			raise ValueError(f"Invalid embedding model {self.llm}")
					 


"""Request Models"""
from .prompt import ReqBodyPromptSystem
from .chat import (ReqBodyChat, ReqBodyAgentChat, ReqBodyAgentPluginsChat,
                    ReqBodyVectorstoreChat, ReqBodyFunctionChat) 
from .retrieval import RequestMultiLoader, RequestDataLoader
from .history import ReqBodyHistory
from .settings import ReqBodySettings


__all__ = [
    'ReqBodyPromptSystem',
    'ReqBodyChat',
    'ReqBodyAgentChat',
    'ReqBodyAgentPluginsChat',
    'ReqBodyVectorstoreChat',
    'ReqBodyFunctionChat',
    'ReqBodyHistory',
    'RequestMultiLoader',
    'RequestDataLoader',
    'ReqBodySettings'
]


from abc import ABC, abstractmethod
from typing import Union, Optional, Dict, Any

class IUserRepo(ABC):

    @abstractmethod
    def find(self, user_id: str = None) -> Optional[Dict[str, Any]]:
        """
        Method to find user data.
        :param user_id: User's ID.
        :return: User data or None.
        """
        pass

    @abstractmethod
    def find_token(self, user_id: str, token_names: Union[str, list], legacy: bool = False) -> Union[str, Dict[str, Any], None]:
        """
        Method to find API token(s).
        :param user_id: User's ID.
        :param token_names: Name(s) of the token(s) to find.
        :param legacy: Flag to use legacy method.
        :return: Token value(s) or None.
        """
        pass



from bson.objectid import ObjectId

from promptengineers.core.interfaces.controllers import IController
from promptengineers.core.interfaces.repos import IUserRepo
from promptengineers.repos.user import UserRepo
from promptengineers.mongo.service import MongoService
from promptengineers.models.request import ReqBodySettings

class SettingsController(IController):
	def __init__(
		self, 
		request = None,
		user_repo: IUserRepo = None, 
		db_name: str = None,
		col_name: str = None
	):
		self.request = request
		self.user_id = getattr(request.state, "user_id", None)
		self.user_repo = user_repo or UserRepo()
		self.settings_service = MongoService(
			host=self.user_repo.find_token(self.user_id, 'MONGO_CONNECTION'),
			db=db_name or self.user_repo.find_token(self.user_id, 'MONGO_DB_NAME'),
			collection=col_name or 'settings'
		)

	##############################################################
	### Index
	##############################################################
	async def index(self, page: int = 1, limit: int = 10):
		result = await self.settings_service.list_docs(
			{'user_id': ObjectId(self.user_id)},
			limit,
			page
		)
		return result

	##############################################################
	### Create
	##############################################################
	async def create(
			self, 
			body: ReqBodySettings, 
			keys: set[str] = {'system', 'temperature', 'retrieval', 'tags',
		  						'functions', 'tools', 'plugins', 'title', 'model', 'stream'}
		):
		body = await self.request.json()
		body = dict((k, body[k]) for k in keys if k in body)
		body['user_id'] = ObjectId(self.user_id)
		result = await self.settings_service.create(dict(body))
		return result

	##############################################################
	### Show
	##############################################################
	async def show(self, id: str):
		result = await self.settings_service.read_one(
			{'_id': ObjectId(id), 'user_id': ObjectId(self.user_id)}
		)
		return result


	##############################################################
	### Update
	##############################################################
	async def update(
		self, 
		id: str, 
		body: ReqBodySettings,
		keys: set[str] = {'system', 'temperature', 'retrieval', 'tags',
		  						'functions', 'tools', 'plugins', 'title', 'model', 'stream'}
	):
		body = await self.request.json()
		body = dict((k, body[k]) for k in keys if k in body)
		result = await self.settings_service.update_one(
			{'_id': ObjectId(id), 'user_id': ObjectId(self.user_id)},
			dict(body)
		)
		return result

	##############################################################
	### Delete
	##############################################################
	async def delete(self, id: str):
		result = await self.settings_service.delete_one(
			{'_id': ObjectId(id), 'user_id': ObjectId(self.user_id)}
		)
		return result


from enum import Enum


class UrlLoaderType(str, Enum):
	GITBOOK = 'gitbook'
	WEB_PAGE = 'web_base'
	WEBSITE = 'website'
	YOUTUBE = 'yt'
	SITEMAP = 'sitemap'
	URLS = 'urls'
	COPY = 'copy'
	ETHEREUM = 'ethereum'
	POLYGON = 'polygon'


class FileLoaderType(str, Enum):
	TEXT = 'txt'
	HTML = 'html'
	MARKDOWN = 'md'
	CSV = 'csv'
	PDF = 'pdf'
